// LCD drive routine for HD44780 16x2 Display
// 2012/09/30 Eleken

#include <inttypes.h>
#include <avr/io.h>
#include <util/delay.h>


#include "lcd.h"

// macros
#define sbi(port, bit)	(port |= _BV(bit))
#define cbi(port, bit)	(port &= ~_BV(bit))

static void LCD_Transfer_Write4(const uint8_t data){
	uint8_t ddr_port_old = DATA_DDR & (~(0x0F << DATA_PIN));

	cbi(RS_PORT, RS_PIN); // set RS
	cbi(RW_PORT, RW_PIN); // write mode
	
	// higher 4 bit
	sbi(E_PORT,  E_PIN );
	DATA_DDR = ddr_port_old | (0x0F << DATA_PIN); // set io port (data bus) to output
	DATA_PORT = (DATA_PORT & DATA_PORT_MASK) | ((data>>4) << DATA_PIN);
	cbi(E_PORT,  E_PIN );
	
	DATA_DDR = ddr_port_old; // houkou register wo modosu
}

// Initialize LCD
void LCD_Init(void){
	_delay_ms(20);

	// Initializing by Instruction
	LCD_Transfer_Write4(0x30); // Function Set
	_delay_ms(10);
	LCD_Transfer_Write4(0x30); // Function Set
	_delay_ms(1);
	LCD_Transfer_Write4(0x30); // Function Set

	_delay_ms(1);
	LCD_Transfer_Write4(0x20); // Function Set (IF = 4bit)
	LCD_Transfer_Write(0x20 | 8, 0); // Function Set (IF = 4bit, 2 Lines, 5x8 Fonts)
	LCD_BusyCheck();
	LCD_Transfer_Write(0x08 | 4, 0); // Display Off
	LCD_BusyCheck();
	LCD_Transfer_Write(0x01, 0); // Display Clear
	LCD_BusyCheck();
	LCD_Transfer_Write(0x04 | 2, 0); // Entry Mode Set
	LCD_BusyCheck();

/*
	// Reset Display with POR
	LCD_Transfer_Write4(0x20); // Function Set (IF = 4bit)
	LCD_Transfer_Write(0x20 | 8, 0); // Function Set (IF = 4bit, 2 Lines, 5x8 Fonts)
	LCD_BusyCheck();
	LCD_Transfer_Write(0x08 | 4, 0); // Display On
	LCD_BusyCheck();
	LCD_Transfer_Write(0x04 | 2, 0); // Entry Mode Set
	LCD_BusyCheck();
*/
}

// 1-byte transfer
void LCD_Transfer_Write(const uint8_t data, const uint8_t rs){
	uint8_t ddr_port_old = DATA_DDR & DATA_PORT_MASK;

	(rs)? sbi(RS_PORT, RS_PIN): cbi(RS_PORT, RS_PIN); // set RS
	cbi(RW_PORT, RW_PIN); // write mode
	
	// higher 4 bit
	sbi(E_PORT,  E_PIN );
	DATA_DDR = ddr_port_old | (0x0F << DATA_PIN); // set io port (data bus) to output
	DATA_PORT = (DATA_PORT & DATA_PORT_MASK) | ((data>>4) << DATA_PIN);
	cbi(E_PORT,  E_PIN );
	
	// lower 4 bit
	sbi(E_PORT,  E_PIN );
	DATA_PORT = (DATA_PORT & DATA_PORT_MASK) | ((data&0x0F) << DATA_PIN);
	cbi(E_PORT,  E_PIN );
	
	DATA_DDR = ddr_port_old; // houkou register wo modosu
}
// 1-byte transfer
void LCD_Transfer_Read(uint8_t* data, const uint8_t rs){
	DATA_DDR = DATA_DDR & DATA_PORT_MASK; // set io port (data bus) to input
	
	uint8_t readdata = 0;

	(rs)? sbi(RS_PORT, RS_PIN): cbi(RS_PORT, RS_PIN); // set RS
	sbi(RW_PORT, RW_PIN); // read mode
	_delay_us(1);
	
	// higher 4 bit
	sbi(E_PORT,  E_PIN );
	_delay_us(1);
	readdata = ((DATA_INPUT >> DATA_PIN) << 4) & 0xF0;
	cbi(E_PORT,  E_PIN );
	
	// lower 4 bit
	sbi(E_PORT,  E_PIN );
	_delay_us(1);
	readdata |= (DATA_INPUT >> DATA_PIN) & 0x0F;
	cbi(E_PORT,  E_PIN );
	
	cbi(RW_PORT, RW_PIN);
	*data = readdata;
}
// Polling until Busy Flag = 0
void LCD_BusyCheck(void){
	uint8_t ret;
	for(;;){
		LCD_Transfer_Read(&ret, 0);
		if(!(ret & 0x80)) break;
	}
	_delay_us(10); // additional delay
}

void LCD_Clear(void){
	LCD_Transfer_Write(0x01, 0);
	LCD_BusyCheck();
}

// Set cursel pos (pos = row*0x40 | column; row:{0,1}, column<40)
void LCD_SetTextPos(const uint8_t pos){
	LCD_Transfer_Write(0x80 | pos, 0);
	LCD_BusyCheck();
}

void LCD_Print(const char* text, uint8_t size){
	uint8_t i;
	for(i=0; i<size; i++){
		LCD_Transfer_Write(text[i], 1);
		LCD_BusyCheck();
//		_delay_us(10); // additional delay: see HD44780 datasheet p.25
	}
}
void LCD_Putc(const char text){
	LCD_Transfer_Write(text, 1);
	LCD_BusyCheck();
}
void LCD_PrintBCD(const uint16_t bcd){
	if(bcd & 0xF000){
		LCD_Putc('0' + ((bcd>>12) & 0x000F));
		LCD_Putc('0' + ((bcd>>8) & 0x000F));
		LCD_Putc('0' + ((bcd>>4) & 0x000F));
		LCD_Putc('0' + ((bcd>>0) & 0x000F));
	}else if(bcd & 0x0F00){
		LCD_Putc('0' + ((bcd>>8) & 0x000F));
		LCD_Putc('0' + ((bcd>>4) & 0x000F));
		LCD_Putc('0' + ((bcd>>0) & 0x000F));
	}else if(bcd & 0x00F0){
		LCD_Putc('0' + ((bcd>>4) & 0x000F));
		LCD_Putc('0' + ((bcd>>0) & 0x000F));
	}else{
		LCD_Putc('0' + ((bcd>>0) & 0x000F));
	}
}

