// Super LED tester for ATMEAG8
// 2012/09/30 Eleken
// For more information: http://www.adachi.ne.jp/users/2sc1815/

// main clock : 8.0 MHz
// Fuses: High bit = 0b11010001
//        Low  bit = 0b10100100
// (1: Not program, 0: Program)

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <util/delay.h>

//#include "main.h"

#include "lcd.h"

// macros
#define sbi(port, bit)	(port |= _BV(bit))
#define cbi(port, bit)	(port &= ~_BV(bit))


// definition of constants

// declare variables

volatile uint8_t  gSwState; // state of switch [0-4] : none, A, B, C, A-long
volatile uint16_t gADCValue1, gADCValue2; // ADCValue, Fixed point, BCD (x.xx[V])
volatile uint8_t  gCurrentSetting1, gCurrentSetting2; // Current settings (0-20: 0-20mA)
volatile uint8_t  gConnection; // Connection state (bit 0,4: isConnect)
volatile uint16_t  gICRValue; // ICR Value (>255)
volatile uint16_t gPowerOnTimer; // Power On Count (div/0.1s)

// eeprom
uint8_t dummy __attribute__((section(".eeprom")));
uint8_t eepCurrent1 __attribute__((section(".eeprom")));
uint8_t eepCurrent2 __attribute__((section(".eeprom")));
uint16_t eepICRValue __attribute__((section(".eeprom")));


// declare functions
static void GetAdcValue(void);


// timer0 : 256us
ISR(TIMER0_OVF_vect){

	static uint16_t t0cnt = 0;
	static uint8_t sw_old;
	static uint8_t longpush_cnt;
	
	// checks switch state ( 0.25us * 400 = 100ms )
	if((t0cnt == 0) && (gSwState == 0)){ 
		uint8_t sw = (PINB & 0x38);
		
		
		if(sw_old == sw){ // if switch state has not changed
			gPowerOnTimer++;
			gSwState = 0;// 
		}else { // switch state has changed
			gPowerOnTimer = 0;
			if(sw == 0x18){//B
				gSwState = 2;
			}else if(sw == 0x28){//C
				gSwState = 3;
			}else if(sw == 0x38){//off
				if(longpush_cnt == 0) gSwState = 0; // longpush?
				else if(longpush_cnt <= 6) gSwState = 1; // < 600ms
				else gSwState = 4;
			}else gSwState = 0;
		}
		if(sw == 0x30){ //A
			longpush_cnt ++;
		}else{
			longpush_cnt = 0;
		}
		sw_old = sw;
	}

	
	if(++t0cnt > 400) t0cnt = 0;
}


void decimal_add(uint8_t* p){
	if(((*p) & 0x0F) == 9) (*p) += 7;
	else (*p) ++;
}
void decimal_dec(uint8_t* p){
	if(((*p) & 0x0F) == 0) (*p) -= 7;
	else (*p) --;
}
void bcd_to_hex(const uint8_t bcd, uint8_t* hex){
	*hex = (bcd>>4)*10 + (bcd&0x0F);
}
// 16-bit hex -> BCD (input range is limited to 0x99)
uint8_t hex_to_bcd(uint8_t hex){
	uint8_t bcd = 0;
	while(hex >= 10){   hex-=10;   bcd+=0x10;}
	bcd += hex;
	return bcd;
}
// 16-bit hex -> BCD (input range is limited to 0x9999)
uint16_t hex16_to_bcd16(uint16_t hex){
	uint16_t bcd = 0;
	while(hex >= 1000){ hex-=1000; bcd+=0x1000;}
	while(hex >= 100){  hex-=100;  bcd+=0x100;}
	while(hex >= 10){   hex-=10;   bcd+=0x10;}
	bcd += hex;
	return bcd;
}

static void GetAdcValue(void){
	uint16_t temp = 0;
	uint8_t  i;

	uint16_t vref; // 2.5V refernce (ADC2)
	uint32_t ch1, ch2;
	
	// AD Convert
	ADMUX  = 0x40 | 2; // Vref = AVCC, MUX = 2
	_delay_us(10);
	temp = 0;
	for(i=0; i<16; i++){
		ADCSRA = 0b11010110; // ADC Start [Note: MUST SET ADIF bit into 1 to clear ADIF bit]
		while(!(ADCSRA & _BV(ADIF))); // Polling until finish
		vref = ADC;
		temp += vref;
	}
	vref = temp >> 4;

	ADMUX  = 0x40 | 0; // Vref = AVCC, MUX = 0
	_delay_us(10);
	temp = 0;
	for(i=0; i<16; i++){
		ADCSRA = 0b11010110; // ADC Start
		while(!(ADCSRA & _BV(ADIF))); // Polling until finish
		ch1 = ADC;
		temp += ch1;
	}
	ch1 = temp >> 4;

	ADMUX  = 0x40 | 1; // Vref = AVCC, MUX = 1
	_delay_us(10);
	temp = 0;
	for(i=0; i<16; i++){
		ADCSRA = 0b11010110; // ADC Start
		while(!(ADCSRA & _BV(ADIF))); // Polling until finish
		ch2 = ADC;
		temp += ch2;
	}
	ch2 = temp >> 4;
	
	// Scale conversion
	ch1 = ((1023-ch1) * 250 / vref);
	ch2 = ((1023-ch2) * 250 / vref);
	
	gADCValue1 = hex16_to_bcd16((uint16_t)ch1); 
		// = hex16_to_bcd16(x); x = ch1/vref*2.5*100 = ch1*250/vref
	gADCValue2 = hex16_to_bcd16((uint16_t)ch2);
		// = hex16_to_bcd16(x); x = ch1/vref*2.5*100 = ch1*250/vref
		
//	gADCValue1 = hex16_to_bcd16((uint16_t)ch2); 
//	gADCValue2 = hex16_to_bcd16((uint16_t)vref); 
}

static void UpdateDisplay(void){
//	1234567890123456
//	CH1: 20mA 3.15V
//	CH2: 20mA  N.C.
//	1: 20mA Vf=3.15V
//	2: 20mA (N.C.)
	uint8_t bcd;
	
	uint8_t offs = 0x40;
	// CH.1
	LCD_SetTextPos(offs + 0);
	LCD_Print("2:   mA Vf= .  V", 16);
	// Current
	bcd = hex_to_bcd(gCurrentSetting1);
	LCD_SetTextPos(offs + 3);
	if(bcd >> 4) LCD_Putc('0' + (bcd >> 4));
	else LCD_Putc(' ');
	LCD_Putc('0' + (bcd & 0x0F));
	
	// Voltage
	if(gConnection & 0x01){ // CH1 Connected
		LCD_SetTextPos(offs + 11);
		LCD_Putc('0' + ((gADCValue1 >> 8) & 0x0F));
		LCD_Putc('.');
		LCD_Putc('0' + ((gADCValue1 >> 4) & 0x0F));
		LCD_Putc('0' + ((gADCValue1 >> 0) & 0x0F));
	}else{
		LCD_SetTextPos(offs + 8);
		LCD_Print("(N.C.)  ", 8);
	}
	
	offs = 0x00;
	// CH.2
	LCD_SetTextPos(offs + 0);
	LCD_Print("1:   mA Vf= .  V", 16);
	// Current
	bcd = hex_to_bcd(gCurrentSetting2);
	LCD_SetTextPos(offs + 3);
	if(bcd >> 4) LCD_Putc('0' + (bcd >> 4));
	else LCD_Putc(' ');
	LCD_Putc('0' + (bcd & 0x0F));
	
	// Voltage
	if(gConnection & 0x10){ // CH21 Connected
		LCD_SetTextPos(offs + 11);
		LCD_Putc('0' + ((gADCValue2 >> 8) & 0x0F));
		LCD_Putc('.');
		LCD_Putc('0' + ((gADCValue2 >> 4) & 0x0F));
		LCD_Putc('0' + ((gADCValue2 >> 0) & 0x0F));
	}else{
		LCD_SetTextPos(offs + 8);
		LCD_Print("(N.C.)  ", 8);
	}
}

static void Init(void){
	
	cli();
	
	// init pins
	DDRC  = 0b00111000;
	PORTC = 0b00100000;
	
	DDRB  = 0b00000111;
	PORTB = 0b00111000;
	
	DDRD  = 0b00001111;
	PORTD = 0b11110000;
	

	// Load EEPROM Settings
	gCurrentSetting1 = eeprom_read_byte(&eepCurrent1);
	gCurrentSetting2 = eeprom_read_byte(&eepCurrent2);
	if((gCurrentSetting1 < 1) || (gCurrentSetting1 > 20)) gCurrentSetting1 = 1;
	if((gCurrentSetting2 < 1) || (gCurrentSetting2 > 20)) gCurrentSetting2 = 1;
	gICRValue = eeprom_read_word(&eepICRValue);
	if((gICRValue < 255) || (gICRValue > 512)) gICRValue = 320;

	
	// init IO registers
	
	// TIMER
	TIMSK  = 0b00000001; // enables TOIE0 (Timer0 overflow interrupt)
	
	TCCR1A = 0b11110010; // enable high speed PWM, Inverted, TOP = ICR1
	TCCR1B = 0b00011001; // clk=clk/1, 
	ICR1   = gICRValue;
	OCR1A  = gICRValue-0;
	OCR1B  = gICRValue-0; 
	TCNT1  = 0;
	
	TCCR0  = 0b00000010; // CS0 = 2 (CLK/8)
	TCNT0  = 0;
	
	// ADC
	ADMUX  = 0x40 | 0; // Vref = AVCC, MUX = 0
	ADCSRA = 0b10000110; // ADCClk = CK/64
	
	LCD_Init();
	
	
	sei();
}

int main(void){

	Init();
	
	LCD_SetTextPos(0);
	LCD_Print("Super LED Tester", 16);
	LCD_SetTextPos(0x40);
	LCD_Print("(C)2012 Eleken  ", 16);
	for(uint8_t i=0; i<4; i++) _delay_ms(100);
	LCD_Clear();
	
	if((PINB & 0x20) == 0){ // Enter Setting Mode (Adjust gICRValue with 10mA current)
		while((PINB & 0x20) == 0);
		
		for(;;){
			// Check switch
			if(gSwState == 1){ // A push
				gSwState = 0;
				eeprom_write_word(&eepICRValue, gICRValue);
				break;
			}else if(gSwState == 3){ // C push
				gICRValue --;
			}else if(gSwState == 2){ // B push
				gICRValue ++;
			}
			gSwState = 0;
			
			// Update display
			LCD_SetTextPos(0);
			LCD_Print("ICR = ", 6);
			uint16_t bcd = hex16_to_bcd16(gICRValue);
			LCD_PrintBCD(bcd);
			
			ICR1   = gICRValue;
			OCR1A  = gICRValue-100;
			OCR1B  = gICRValue-100;
		}
	}
	
	gConnection = 0;
	GetAdcValue();
	
	for(;;){
		// Check switch
		
		if(gSwState == 1){ // A push
			gSwState = 0;
		}else if(gSwState == 3){ // C push
			gSwState = 0;
			
			if(gCurrentSetting1 > 10) gCurrentSetting1 -= 2;
			else if(gCurrentSetting1 > 1) gCurrentSetting1 -= 1;
			if(gCurrentSetting2 > 10) gCurrentSetting2 -= 2;
			else if(gCurrentSetting2 > 1) gCurrentSetting2 -= 1;
		}else if(gSwState == 2){ // B push
			gSwState = 0;
			
			if(gCurrentSetting1 < 10) gCurrentSetting1 += 1;
			else if(gCurrentSetting1 < 20) gCurrentSetting1 += 2;
			if(gCurrentSetting2 < 10) gCurrentSetting2 += 1;
			else if(gCurrentSetting2 < 20) gCurrentSetting2 += 2;
		}else if((gSwState == 4) || (gPowerOnTimer > 3000)){ // A long push or power-on timer is over 5min.
			gSwState = 0;
			
			// Save EEPROM Settings
			eeprom_write_byte(&eepCurrent1, gCurrentSetting1);
			eeprom_write_byte(&eepCurrent2, gCurrentSetting2);
			
			cbi(PORTC, PC5); // power off
		}
		
		// Update PWM Value
		if(gConnection & 0x01){ // CH1 Connected
			OCR1A = gICRValue - (uint16_t)(gCurrentSetting1*10);
			if(gADCValue1 > 0x450){ // Detect Disconnection
				gConnection &= ~0x01;
				
				OCR1A = gICRValue;
				for(uint8_t i=0; i<2; i++) _delay_ms(100);
			}
		}else{
			if(gADCValue1 < 0x300){ // Detect Connection
				gConnection |= 0x01;
				
				OCR1A = gICRValue - (uint16_t)(gCurrentSetting1*10);
				for(uint8_t i=0; i<2; i++) _delay_ms(100);
			}
		}
		if(gConnection & 0x10){ // CH2 Connected
			OCR1B = gICRValue - (uint16_t)(gCurrentSetting2*10);
			if(gADCValue2 > 0x450){ // Detect Disconnection
				gConnection &= ~0x10;
				
				OCR1B = gICRValue;
				for(uint8_t i=0; i<2; i++) _delay_ms(100);
			}
		}else{
			if(gADCValue2 < 0x300){ // Detect Connection
				gConnection |= 0x10;
				
				OCR1B = gICRValue - (uint16_t)(gCurrentSetting2*10);
				for(uint8_t i=0; i<2; i++) _delay_ms(100);
			}
		}
		
		GetAdcValue();
		
		UpdateDisplay();
		_delay_ms(100);
	}

	return 0;
}

